/*
 * File........: main.c
 * Author......: Johannes Heimansberg (wejp.k.vu)
 *               Based on hid-custom-rq example by Christian Starkjohann
 * Date created: 20120215
 * License.....: GNU GPLv2
 */

#define LED_PORT_DDR    DDRB
#define LED_PORT_OUTPUT PORTB
#define COLOR_1_BIT     3
#define COLOR_2_BIT     4

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>  /* for sei() */
#include <util/delay.h>     /* for _delay_ms() */
#include <avr/pgmspace.h>   /* required by usbdrv.h */
#include "usbdrv.h"
#include "oddebug.h"        /* This is also an example for using debug macros */
#include "requests.h"       /* The custom request numbers we use */

PROGMEM const char usbHidReportDescriptor[22] = { /* USB report descriptor */
    0x06, 0x00, 0xff,              // USAGE_PAGE (Generic Desktop)
    0x09, 0x01,                    // USAGE (Vendor Usage 1)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x26, 0xff, 0x00,              //   LOGICAL_MAXIMUM (255)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x95, 0x01,                    //   REPORT_COUNT (1)
    0x09, 0x00,                    //   USAGE (Undefined)
    0xb2, 0x02, 0x01,              //   FEATURE (Data,Var,Abs,Buf)
    0xc0                           // END_COLLECTION
};
/* The descriptor above is a dummy only, it silences the drivers. The report
 * it describes consists of one byte of undefined data.
 * We don't transfer our data through HID reports, we use custom requests
 * instead.
 */
static uint8_t count = 0;
static volatile uint8_t led1v = 0;
static volatile uint8_t led2v = 0;
static uint8_t led1 = 0;
static uint8_t led2 = 0;
static volatile uint8_t button_mode = 0; /* Default: non-USB mode */
static uint16_t btn_cnt = 0;
static volatile uint8_t button_pressed = 0;


usbMsgLen_t usbFunctionSetup(uchar data[8])
{
	usbRequest_t *rq = (usbRequest_t *)data;
	static uchar  msg[1];
	uint8_t       res_size = 0; /* default for not implemented requests: return no data back to host */

    if ((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_VENDOR) {
		switch(rq->bRequest) {
			case CUSTOM_RQ_SET_LED1:
				led1v = rq->wValue.bytes[0];
				break;
			case CUSTOM_RQ_SET_LED2:
				led2v = rq->wValue.bytes[0];
				break;
			case CUSTOM_RQ_CONFIGURE_BUTTON:
				button_mode = rq->wValue.bytes[0];
				break;
			case CUSTOM_RQ_READ_BUTTON:
				msg[0] = button_pressed ? 'P' : 'N';
				button_pressed = 0;
				usbMsgPtr = msg;
				res_size = 1; /* We return one character */
				break;
		}
    } else {
        /* calss requests USBRQ_HID_GET_REPORT and USBRQ_HID_SET_REPORT are
         * not implemented since we never call them. The operating system
         * won't call them either because our descriptor defines no meaning.
         */
    }
    return res_size;
}

static void calibrateOscillator(void)
{
	uchar       step = 128;
	uchar       trialValue = 0, optimumValue;
	int         x, optimumDev, targetValue = (unsigned)(1499 * (double)F_CPU / 10.5e6 + 0.5);
 
	/* do a binary search: */
	do {
		OSCCAL = trialValue + step;
		x = usbMeasureFrameLength(); /* proportional to current real frequency */
		if(x < targetValue)          /* frequency still too low */
			trialValue += step;
		step >>= 1;
	} while (step > 0);
	/* We have a precision of +/- 1 for optimum OSCCAL here */
	/* now do a neighborhood search for optimum value */
	optimumValue = trialValue;
	optimumDev = x; /* this is certainly far away from optimum */
	for (OSCCAL = trialValue - 1; OSCCAL <= trialValue + 1; OSCCAL++) {
		x = usbMeasureFrameLength() - targetValue;
		if (x < 0)
			x = -x;
		if (x < optimumDev) {
			optimumDev = x;
			optimumValue = OSCCAL;
		}
	}
	OSCCAL = optimumValue;
}

void usbEventResetReady(void)
{
	cli(); /* usbMeasureFrameLength() counts CPU cycles, so disable interrupts. */
	calibrateOscillator();
	sei();
	eeprom_write_byte(0, OSCCAL); /* store the calibrated value in EEPROM */
}

void update_leds()
{
	if (++count == 0) {
		led1 = led1v;
		led2 = led2v;
		/* Turn LEDs on */
		if (led1 > 0) PORTB |= (1 << COLOR_1_BIT);
		if (led2 > 0) PORTB |= (1 << COLOR_2_BIT);
	}
	/* Turn LEDs off */
	if (count == led1) PORTB &= ~(1 << COLOR_1_BIT);
	if (count == led2) PORTB &= ~(1 << COLOR_2_BIT);
}

int main(void)
{
	wdt_enable(WDTO_1S);
	/* Even if you don't use the watchdog, turn it off here. On newer devices,
	 * the status of the watchdog (on/off, period) is PRESERVED OVER RESET!
	 */
	/* RESET status: all port bits are inputs without pull-up.
	 * That's the way we need D+ and D-. Therefore we don't need any
	 * additional hardware initialization.
	 */

	usbInit();
	usbDeviceConnect();
	LED_PORT_DDR |= _BV(COLOR_1_BIT); /* make the LED bit an output */
	LED_PORT_DDR |= _BV(COLOR_2_BIT); /* make the LED bit an output */

	/* Configure input pin with internal pull-up for the button */
	DDRB  &= 0b11111101; /* Just zero bit 1, leave everything else alone */
	PORTB |= 0b00000010; /* Just set bit 1, leave everything else alone  */

	sei();

	for(;;) {
		/* Check for button press... (button debounce: btn_cnt must be *
		 * 0 for the button press to be detected)                      */
		if (!(PINB & 0b00000010) && !btn_cnt) {
			button_pressed = 1;
			btn_cnt = 65535;
		} else if (btn_cnt) {
			btn_cnt--;
		}
		if (button_pressed) {
			switch (button_mode) {
				case BUTTON_NORMAL:
					if (led1v) {
						led1v = 0;
					} else {
						led1v = 255;
						if (led2v) led2v = 0; else led2v = 255;
					}
					button_pressed = 0;
					break;
				case BUTTON_OFF_BUTTON:
					led1v = 0;
					led2v = 0;
					button_pressed = 0;
					break;
				default:
				case BUTTON_USB:
					break;
			}
		}

		update_leds();
		wdt_reset();
		usbPoll();
	}
	return 0;
}
