/* 
 * File: spokausb.c  Created: 120215
 *
 * Copyright (c) 2012 Johannes Heimansberg (wejp.k.vu)
 * 
 * Description: Control program/daemon for the USB-enabled SPÖKA light
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; version 2 of
 * the License. See the file COPYING for details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <syslog.h>
#include <usb.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/stat.h>
#include <sys/sem.h>
#include <sys/types.h>
#include <errno.h>
#include "opendevice.h"
#include "daemon.h"
#include "messagequeue.h"

#include "../firmware/requests.h"
#include "../firmware/usbconfig.h"

#define VERSION "0.1.0"

static void usage(char *name)
{
	fprintf(stderr, "spokausb " VERSION "\n");
	fprintf(stderr, "Usage:\n");
	fprintf(stderr, "  %s [option] [...]\n", name);
	fprintf(stderr, "Options:\n");
	fprintf(stderr, " -d  .................................... start daemon\n");
	fprintf(stderr, " --led{lednum}={brightness}  ............ set brightness of LED\n");
	fprintf(stderr, " --led{lednum}+={brightness increase}  .. increase brightness of LED\n");
	fprintf(stderr, " --led{lednum}-={brightness decrease}  .. decrease brightness of LED\n");
	fprintf(stderr, " --button=<normal|user|disable>.......... configure button mode\n");
	fprintf(stderr, " --fade   ............................... slowly cycle through all colors\n");
	fprintf(stderr, " --flash  ............................... flash repeatedly\n");
	fprintf(stderr, " --freeze  .............................. stop cycling through colors\n");
	fprintf(stderr, " --read-button  ......................... read button state\n");
	fprintf(stderr, " --read-led{lednum}  .................... reade LED state\n");
	fprintf(stderr, " --help/-h  ............................. this help\n");
}

static int setup_message_queue(key_t key, int flags)
{
	int res = msgget(key, flags);
	if (res < 0) {
		printf("Error setting up message queue. :(\n");
		res = -1;
	}
	return res;
}

static int client(int argc, char **argv)
{
	int     server_id, client_id, res, i;
	Message c2s, s2c;

	for (i = 1; i < argc; i++) {
		if ((server_id = setup_message_queue(MAGIC_KEY, 0)) >= 0) {
			if ((client_id = setup_message_queue(IPC_PRIVATE, 0666 | IPC_CREAT)) >= 0) {
				c2s.priority = 2;
				memset(c2s.message, 0, MESSAGE_LENGTH);
				snprintf(c2s.message, MESSAGE_LENGTH, "%d|%s", client_id, argv[i]);
				if ((res = msgsnd(server_id, &c2s, MESSAGE_LENGTH, 0)) != -1) {
					if (strncmp(argv[i], "--read", 6) == 0) {
						int i;
						errno = 0;
						for (res = -1, i = 10; i > 0 && res < 0; i--) {
							res = msgrcv(client_id, &s2c, MESSAGE_LENGTH, 0, IPC_NOWAIT);
							if (res < 0) usleep(50000);
						}
						if (res > 0)
							printf("%s", s2c.message);
						else
							printf("No reply received");
					}
					errno = 0;
				}
				msgctl(client_id, IPC_RMID, NULL);
			}
		}
	}
	return 0;
}

static int led_set_brightness(usb_dev_handle *handle, unsigned char lednum, unsigned char brightness)
{
	int led = CUSTOM_RQ_SET_LED1;
	if (lednum == 2) led = CUSTOM_RQ_SET_LED2;
	/*syslog(LOG_NOTICE, "Setting brightness for LED %d to value %d...\n", lednum, brightness);*/
	return usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
	                       led, brightness, 0, NULL, 0, 5000);
}

static int button_read(usb_dev_handle *handle, int *result)
{
	static char usb_buffer[4];
	int res;
	res = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
	                      CUSTOM_RQ_READ_BUTTON, 0, 0, (char *)usb_buffer, sizeof(usb_buffer), 5000);
	*result = usb_buffer[0] == 'P' ? 1 : 0;
	return res;	
}

static int button_set_mode(usb_dev_handle *handle, int mode)
{
	syslog(LOG_NOTICE, "Configuring SPOKA button for mode %d...\n", mode);
	return usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
	                       CUSTOM_RQ_CONFIGURE_BUTTON, mode, 0, NULL, 0, 5000);
}

typedef enum {
	SPOKA_NORMAL, SPOKA_FLASH, SPOKA_FADE
} SPOKA_Mode;

static int server_send_reply(int client_id, char *msg)
{
	Message s2c;
	int     res = 0;
	s2c.priority = 1;
	snprintf(s2c.message, MESSAGE_LENGTH, "%s", msg);
	if (msgsnd(client_id, &s2c, MESSAGE_LENGTH, 0) != 0) {
		syslog(LOG_NOTICE, "Could not send message to client. :(\n");
		res = -1;
	}
	return res;
}

static int stop_daemon = 0;

static void sig_handler(int sig)
{
	syslog(LOG_NOTICE, "Stopping daemon...\n");
	stop_daemon = 1;
}

#define FLASH_DELAY 2
#define FADE_DELAY  1

static int server_main(void)
{
	int                 server_id, res;
	usb_dev_handle     *handle = NULL;
	const unsigned char rawVid[2] = {USB_CFG_VENDOR_ID}; 
	const unsigned char rawPid[2] = {USB_CFG_DEVICE_ID};
	char                vendor[]  = {USB_CFG_VENDOR_NAME, 0};
	char                product[] = {USB_CFG_DEVICE_NAME, 0};
	int                 vid, pid;
	SPOKA_Mode          spoka_mode = SPOKA_NORMAL;
	unsigned char       brightness[2];
	int                 user_readable_button = 0, button_disabled = 0;

	server_id = setup_message_queue(MAGIC_KEY, 0666 | IPC_CREAT);
	usb_init();
	syslog(LOG_NOTICE, "Daemon has been started.\n");
	vid = rawVid[1] * 256 + rawVid[0];
	pid = rawPid[1] * 256 + rawPid[0];

	brightness[0] = brightness[1] = 0;

	for (; server_id >= 0 && !stop_daemon;) { /* USB device reconnect loop */
		static int cnt = FLASH_DELAY;
		int        button_pressed = 0;
		while (usbOpenDevice(&handle, vid, vendor, pid, product, NULL, NULL, NULL) != 0) {
			syslog(LOG_WARNING, "Could not find USB device \"%s\" with vid=0x%x pid=0x%x\n", product, vid, pid);
			usleep(10000000);
		}
		syslog(LOG_NOTICE, "Connected to %s device [%x:%x].\n", product, vid, pid);
		led_set_brightness(handle, 1, 0);
		led_set_brightness(handle, 2, 0);

		for (; !stop_daemon;) { /* main loop */
			Message c2s;
			res = msgrcv(server_id, &c2s, MESSAGE_LENGTH, 0, IPC_NOWAIT);
			if (errno == ENOMSG) {
				errno = 0;
				usleep(300000);
			}

			if (!button_pressed && !button_disabled) {
				int tmp;
				button_read(handle, &tmp);
				if (tmp) button_pressed = 1;
			}

			if (button_pressed && !user_readable_button) {
				switch (spoka_mode) {
					case SPOKA_FADE:
						syslog(LOG_NOTICE, "Color %d/%d frozen by button press!\n", brightness[0], brightness[1]);
						spoka_mode = SPOKA_NORMAL;
						button_pressed = 0;
						break;
					case SPOKA_NORMAL:
						if (brightness[0] || brightness[1]) { /* light is turned on */
							brightness[0] = brightness[1] = 0;
							led_set_brightness(handle, 1, brightness[0]);
							led_set_brightness(handle, 2, brightness[1]);
						} else { /* light is turned off */
							spoka_mode = SPOKA_FADE;
							cnt = FADE_DELAY;
							brightness[0] = brightness[1] = 100;
							button_set_mode(handle, BUTTON_USB);
						}
						button_pressed = 0;
						break;
					default:
						break;
				}
			}

			if (res >= 0) {
				int   client_id = -1, len = 0;
				char *tmp = NULL, *msg = NULL;

				len = strlen(c2s.message);
				if (len > 0) {
					tmp = malloc(len+1);
					if (tmp) {
						int i;
						memcpy(tmp, c2s.message, len+1);
						for (i = 0; i < len; i++) {
							if (tmp[i] == '|') {
								tmp[i] = '\0';
								msg = tmp+i+1;
								client_id = atoi(tmp);
								/*syslog(LOG_NOTICE, "Client: %d Message: '%s'\n", client_id, msg);*/
								break;
							}
						}
					}
				}

				if (strncmp(msg, "--led1", 6) == 0 || strncmp(msg, "--led2", 6) == 0) {
					int lednum = msg[5] - '1' + 1;
					if (msg[6] == '=') {
						/* Set new brightness value */
						int brightness_value = atoi(msg+7);

						spoka_mode = SPOKA_NORMAL;
						if (lednum > 0 && lednum <= 2 && brightness >= 0) {
							brightness[lednum-1] = brightness_value;
							if (led_set_brightness(handle, lednum, brightness[lednum-1])) {
								server_send_reply(client_id, "error");
								break;
							}
						}
					} else if ((msg[6] == '+' || msg[6] == '-') && msg[7] == '=') {
						/* Increase/decrease brightness value */
						int brightness_change = atoi(msg+8);
						int sign   = msg[6] == '-' ? -1 : 1;
						int lednum = msg[5] - '1' + 1;

						spoka_mode = SPOKA_NORMAL;
						if (lednum > 0 && lednum <= 2 && brightness >= 0) {
							unsigned char old = brightness[lednum-1];

							brightness[lednum-1] += sign * brightness_change;
							if      (sign == -1 && brightness[lednum-1] > old) brightness[lednum-1] = 0;
							else if (sign ==  1 && old > brightness[lednum-1]) brightness[lednum-1] = 255;
							if (led_set_brightness(handle, lednum, brightness[lednum-1]))
								break;
						}
					}
				} else if (strncmp(msg, "--button=", 9) == 0) {
					user_readable_button = 0;
					button_disabled = 0;
					if (strcmp(msg+9, "user") == 0) {
						user_readable_button = 1;
					} else if (strcmp(msg+9, "disable") == 0) {
						button_disabled = 1;
					}
					if (button_set_mode(handle, BUTTON_USB)) {
						break;
					}
				} else if (strcmp(msg, "--read-button") == 0) {
					char tmp[64];
					snprintf(tmp, 63, "%d", button_pressed);
					server_send_reply(client_id, tmp);
					button_pressed = 0;
				} else if (strncmp(msg, "--read-led", 10) == 0) {
					int lednum = msg[10] - '1' + 1;
					/* Read brightness value */
					if (lednum > 0 && lednum <= 2) {
						char tmp[64];
						snprintf(tmp, 63, "%d", brightness[lednum-1]);
						server_send_reply(client_id, tmp);
					} else {
						server_send_reply(client_id, "error");
					}
				} else if (strcmp(msg, "--fade") == 0) {
					spoka_mode = SPOKA_FADE;
					cnt = FADE_DELAY;
					brightness[0] = brightness[1] = 100;
					button_set_mode(handle, BUTTON_USB);
					button_pressed = 0;
				} else if (strcmp(msg, "--flash") == 0) {
					spoka_mode = SPOKA_FLASH;
					button_set_mode(handle, BUTTON_USB);
					button_pressed = 0;
				} else if (strcmp(msg, "--freeze") == 0) {
					spoka_mode = SPOKA_NORMAL;
				}
				if (tmp) free(tmp);
			}

			switch (spoka_mode) {
				case SPOKA_NORMAL:
					break;
				case SPOKA_FADE: {
					int led;
					static int dir[2] = { 1, 1 };

					if (cnt == 0) {
						for (led = 0; led <= 1; led++) {
							int prev = brightness[led];
							brightness[led] += dir[led] * (1 + led) * 2;
							if (dir[led] == 1 && brightness[led] < prev) {
								brightness[led] = prev;
								dir[led] = -1;
							}
							if (dir[led] == -1 && brightness[led] > prev) {
								brightness[led] = prev;
								dir[led] = 1;
							}
							if (led_set_brightness(handle, led+1, brightness[led])) break;
						}
						cnt = FADE_DELAY;
					}
					cnt--;
					break;
				}
				case SPOKA_FLASH: {
					int           led = 0;
					unsigned char flash_brightness;

					if (cnt == 0) {
						for (flash_brightness = 255; flash_brightness > 0; flash_brightness -= 5) {
							if (led_set_brightness(handle, led+1, flash_brightness)) break;
							usleep(10000);
						}
						cnt = FLASH_DELAY;
					}
					usleep(200000);
					cnt--;
					break;
				}
			}
		}
		usb_close(handle);
		usleep(1000000);
	}
	msgctl(server_id, IPC_RMID, NULL);
	return 0;
}

static int server(int argc, char **argv)
{
	int sem_id;

	printf("Starting SPOKA daemon...\n");
	if ((sem_id = semget(MAGIC_KEY+1, 0, IPC_PRIVATE)) < 0) {
		umask(0);
		if ((sem_id = semget(MAGIC_KEY+1, 1, IPC_CREAT | IPC_EXCL | 0666)) >= 0) {
			signal(SIGINT,  sig_handler);
			signal(SIGTERM, sig_handler);

			daemonize("spoka", LOG_LOCAL0);
			server_main();
			semctl(sem_id, 0, IPC_RMID, 0);
			syslog(LOG_NOTICE, "Daemon has been stopped.\n");
			closelog();
		} else {
			printf("ERROR: Unable to create semaphore.\n");
		}
	} else {
		printf("ERROR: It looks like the server is already running.\n");
	}
	return 0;
}

int main(int argc, char **argv)
{
	int res = 1;

	if (argc == 2 && strcmp(argv[1], "-d") == 0)
		res = server(argc, argv);
	else if (argc >= 2 && strcmp(argv[1], "--help") != 0 && strcmp(argv[1], "-h") != 0)
		res = client(argc, argv);
	else
		usage(argv[0]);
	return res;
}
